/*
 * 	Description: The class tells if a given set of attributes along with their Functional
 *               dependencies are in BCNF or not.
 *  Version: 1.0
 *  Authors: SJSU Students
 */


import java.io.BufferedReader;
import java.io.FileReader;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

public class BoyceCoddNormalForm {

    static public void main(String[] args){

        if(args.length != 1){
            System.out.println("ERROR: provide file name in command line");
        }
        // Read args to get the file
        Set<FunctionalDependency> fds = new HashSet<>();
        try(BufferedReader in = new BufferedReader(new FileReader(args[0]))){
            String line = in.readLine();
            while(line != null){
                fds.add(new FunctionalDependency(line));
                line = in.readLine();
            }
        } catch(Exception e){
            e.printStackTrace();
        }
        Set<Integer> attributes = new HashSet<>();
        int max = 0;
        for(FunctionalDependency fd : fds){
            for(Integer a : fd.getLeft()){
                max = Math.max(a, max);
            }
            for(Integer a : fd.getRight()){
                max = Math.max(a, max);
            }
        }
        for(int i = 1; i <= max; i++){
            attributes.add(i);
        }

        // Populate the original relation
        Relation originalRelation = new Relation(fds, attributes);

        // Call relation.getBcnfDecomposition
        Set<Relation> decomposition = originalRelation.getBcfnDecomposition();

        // Print the resulting relations
        for(Relation relation : decomposition){
            System.out.println(relation);
        }

    }

    static public class Relation{
        private Set<FunctionalDependency> fd;
        private Set<Integer> attributes;

        /**
         * Constructor
         * @param fd set of functional dependencies
         * @param attributes attributes of the relation
         */
        public Relation (Set<FunctionalDependency> fd, Set<Integer> attributes){
            this.fd = fd;
            this.attributes = attributes;
        }

        /**
         * get the set of functional dependencies
         * @return set of functional dependencies
         */
        public Set<FunctionalDependency> getDependencies(){
            return Set.copyOf(fd);
        }

        /**
         * Get the set of attributes of the relation
         * @return
         */
        public Set<Integer> getAttributes(){
            return Set.copyOf(attributes);
        }

        /**
         * Goes through the list of functional dependencies and removes the trivial ones
         */
        public void removeTrivialFds(){
            for(FunctionalDependency funcDep : fd){
                for(Integer attribute : funcDep.getRight()){
                    if(funcDep.getLeft().contains(attribute)){
                        funcDep.getRight().remove(attribute);
                    }
                }
            }
            for(Iterator<FunctionalDependency> it = fd.iterator(); it.hasNext();){
                FunctionalDependency functionalDependency = it.next();
                if(functionalDependency.getRight().isEmpty()){
                    it.remove();
                }
            }
        }

        /**
         * Goes through the list of functional dependencies and removes the trivial ones
         */
        public static Set<FunctionalDependency> removeTrivialFds2(Set<FunctionalDependency> fds){
            Set<FunctionalDependency> fd = new HashSet<>(fds);
        	for(FunctionalDependency funcDep : fd){
                for(Integer attribute : funcDep.getRight()){
                    if(funcDep.getLeft().contains(attribute)){
                        funcDep.getRight().remove(attribute);
                    }
                }
            }
            for(Iterator<FunctionalDependency> it = fd.iterator(); it.hasNext();){
                FunctionalDependency functionalDependency = it.next();
                if(functionalDependency.getRight().isEmpty()){
                    it.remove();
                }
            }
            return fd;
        }


        /**
         * goes through the functional dependencies in fd and removes the trivials ones
         * @param fds set of functional dependencies
         * @return set with no trivial functional dependencies
         */
        static public Set<FunctionalDependency> removeTrivialFds(Set<FunctionalDependency> fds){
            Set<FunctionalDependency> fd = new HashSet<>(fds);
            for(FunctionalDependency funcDep : fd){
                for(Integer attribute : funcDep.getRight()){
                    if(funcDep.getLeft().contains(attribute)){
                        funcDep.getRight().remove(attribute);
                    }
                }
            }
            for(Iterator<FunctionalDependency> it = fd.iterator(); it.hasNext();){
                FunctionalDependency functionalDependency = it.next();
                if(functionalDependency.getRight().isEmpty()){
                    it.remove();
                }
            }

            return fd;
        }

        /**
         * Check if a relation is in BCNF
         * @return true if the relation is in BCNF
         */
        public boolean isInBCNF(){
            boolean result = true;
            this.removeTrivialFds();
            for(FunctionalDependency dependency : fd){
                Set<Integer> lhs = dependency.getLeft();
                Set<Integer> closure = getClosureOf(lhs);
                if(!attributes.equals(closure)){
                    result = false;
                    break;
                }
            }
            return result;
        }

        /**
         * Get the BCNF decomposition of the relation
         * @return a set of relation all in BCNF
         */
        public Set<Relation> getBcfnDecomposition(){
            // If already in BCNF, return;
            Set<Relation> result = new HashSet<>();
            if(this.isInBCNF()){
                result.add(this);
                return result;
            }

            // Get a Fd that violates BCNF
            FunctionalDependency functionalDependency = fd.iterator().next();
            Set<Integer> closure = getClosureOf(functionalDependency.getLeft());
            Set<Integer> rel2attributes = new HashSet<>();
            rel2attributes.addAll(this.attributes);
            rel2attributes.removeAll(closure);
            rel2attributes.addAll(functionalDependency.getLeft());

            // Project the dependencies
            Set<FunctionalDependency> fdRel1 = projectFD(closure);
            Set<FunctionalDependency> fdRel2 = projectFD(rel2attributes);

            Relation relation1 = new Relation(fdRel1, closure);
            Relation relation2 = new Relation(fdRel2, rel2attributes);
            result.addAll(relation1.getBcfnDecomposition());
            result.addAll(relation2.getBcfnDecomposition());
            return result;
        }

        /**
         * Project the functional dependencies of that relation onto the attributes in R1
         * @param  R1 set of attributes to project the functional dependencies onto
         * @return the set of projected functional dependencies
         */
        public Set<FunctionalDependency> projectFD( Set<Integer> R1){
            Set<FunctionalDependency> T = new HashSet<>();
            Set<Set<Integer>> subsets = powerSet(R1);
            for(Set<Integer> X : subsets){
                Set<Integer> Xclosure = getClosureOf(X);
                Set<FunctionalDependency> nonTrivialFds = Relation.removeTrivialFds(fd);
                for(FunctionalDependency functionalDependency : nonTrivialFds){
                    if(X.containsAll(functionalDependency.getLeft())
                            && Xclosure.containsAll(functionalDependency.getRight())
                            && R1.containsAll(functionalDependency.getRight())){
                        T.add(functionalDependency);
                    }
                }
            }
            return Relation.getMinimalDependency(R1, T);
        }

        /**
         * Get the set of minimal dependencies given attributes and a set of functional dependencies
         * @param R1 set of attributes
         * @param fd set of functional dependencies to minimize
         * @return set of minimal
         */
        static public Set<FunctionalDependency> getMinimalDependency(Set<Integer> R1, Set<FunctionalDependency> fd){
    		Set<FunctionalDependency> result = new HashSet<>(fd);
        	Set<FunctionalDependency> fds = removeTrivialFds2(result);
        	Set<FunctionalDependency> fds2 = splitFunctionalDependencies(fds);

        	//Step 3: remove unnecessary FD's and left side attributes
    		int count = 1;
    		while(count > 0){
    			count = removeUnnecessaryLeftSide(fds2) + removeUnnecessaryEntireFD(fds2);
    		}

    		return fds2;

        }

        public static int removeUnnecessaryEntireFD(Set<FunctionalDependency> fds){
    		int count = 0;
    		while(true){
    			FunctionalDependency toRemove = null;
    			boolean found = false;
    			for(FunctionalDependency fd : fds){
    				Set<FunctionalDependency> remaining = new HashSet<>(fds);
    				remaining.remove(fd);
    				if(equivalent(remaining, fds)){
    					++count;
    					found = true;
    					toRemove = fd;
    					break;
    				}
    			}
    			if(!found){
    				break;
    			}
    			else{
    				fds.remove(toRemove);
    			}
    		}

    		return count;
    	}

        public static int removeUnnecessaryLeftSide(Set<FunctionalDependency> fds){
    		int count = 0;
    		while(true){
    			boolean found = false;
    			FunctionalDependency toRemove = null;
    			FunctionalDependency toAdd = null;
    			int loop = 0;
    			for(FunctionalDependency fd : fds){
    				Set<Integer> left = fd.getLeft();
    				Set<Integer> right = fd.getRight();
    				if(left.size() > 1){
    					for (Integer a : left) {
    						Set<Integer> remaining = new HashSet<>(left);
    						remaining.remove(a);
    						Set<FunctionalDependency> alternative = new HashSet<>(fds);
    						alternative.remove(fd);

    						toAdd = new FunctionalDependency(remaining,right);
    						alternative.add(toAdd);
    						if (equivalent(alternative, fds)) {
    							found = true;
    							toRemove = fd;
    							++count;
    							break;
    						}
    					}
    				}

    				if(found){
    					break;
    				}
    				++loop;
    			}
    			if(found){
    				fds.remove(toRemove);
    				fds.add(toAdd);
    			}
    			if(loop == fds.size()){
    				break;
    			}
    		}
    		return count;
    	}


        public static boolean equivalent(Set<FunctionalDependency> a, Set<FunctionalDependency> b){
    		Set<Integer> names = new HashSet<>();
    		for(FunctionalDependency fd : a){
    			names.addAll(fd.getLeft());
    			names.addAll(fd.getRight());
    		}
    		for(FunctionalDependency fd : b){
    			names.addAll(fd.getLeft());
    			names.addAll(fd.getRight());
    		}

    		Set<Set<Integer>> powerset = reducedPowerSet(names);
    		for(Set<Integer> sa : powerset){
    			Set<Integer> closureInA = closure(sa, a);
    			Set<Integer> closureInB = closure(sa, b);
    			if(!closureInA.equals(closureInB)){
    				return false;
    			}
    		}
    		return true;
    	}

        public static Set<Set<Integer>> reducedPowerSet(Set<Integer> originalSet){
    		Set<Set<Integer>> result = powerSet(originalSet);
    		result.remove(new HashSet<Integer>());
    		return result;
    	}

        public static Set<Integer> closure(Set<Integer> attrs, Set<FunctionalDependency> fds){
    		Set<Integer> result = new HashSet<>(attrs);

    		boolean found = true;
    		while(found){
    			found = false;
    			for(FunctionalDependency fd : fds){
    				if(result.containsAll(fd.left) && !result.containsAll(fd.right)){
    					result.addAll(fd.right);
    					found = true;
    				}
    			}
    		}

    		return result;
    	}


        /**
         * Get all the subsets of a set
         * @param attributes set of attributes
         * @return set of subsets
         */
        static Set<Set<Integer>> powerSet(Set<Integer> attributes){
            Set<Set<Integer>> result = new HashSet<>();
            if (attributes == null || attributes.isEmpty()){
                return result;
            }
            result.add(attributes);
            Iterator<Integer> it = attributes.iterator();
            while(it.hasNext()){
                Set<Integer> temp = new HashSet<>(attributes);
                temp.remove(it.next());
                result.add(temp);
                result.addAll(powerSet(temp));
            }
            return result;
        }

        /**
         * Get the closure of a a set of attributes
         * @param a set of attributes
         * @return closure of a given the relation's functional dependencies
         */
        public Set<Integer> getClosureOf(Set<Integer> a){

        	//Local Variables
        	Set<Integer> X = new HashSet<>(a);
            Set<Integer> Xold = null;

            if(!this.attributes.containsAll(a)){
                throw new IllegalArgumentException( a + " is not an attribute of the relation.");
            }
            this.removeTrivialFds();

            // makes all the fds on right to be single.
            splitFunctionalDependencies();

            while (!X.equals(Xold)) {
            	Xold = new HashSet<>(X);
            	for(FunctionalDependency dependency : fd) {
            		if(X.containsAll(dependency.left) && !X.contains(dependency.right)) {
            			X.addAll(dependency.right);
            		}
            	}
            }

            return X;
        }

        /**
         * Split the relation's functional dependency so the the RHS has only one attribute
         */
        public void splitFunctionalDependencies(){
            Set<FunctionalDependency> splitFds = new HashSet<FunctionalDependency>();
            for(FunctionalDependency dependency : fd){
                if(dependency.right.size() > 1) {
                    for(Integer rightValue : dependency.getRight()) {
                        FunctionalDependency dep = new FunctionalDependency();
                        dep.left = dependency.left;
                        Set<Integer> rightval = new HashSet<Integer>();
                        rightval.add(rightValue);
                        dep.right = rightval;
                        splitFds.add(dep);
                    }
                }else {
                    splitFds.add(dependency);
                }
            }
            this.fd = splitFds;
        }

        /**
         * Given a set of functional dependencies, return an equivalent set where all the RHS have only one attribute
         * @param fd set of functional dependencies
         * @return splited set of functional dependencies
         */
        static Set<FunctionalDependency> splitFunctionalDependencies(Set<FunctionalDependency> fd){
            Set<FunctionalDependency> splitFds = new HashSet<FunctionalDependency>();
            for(FunctionalDependency dependency : fd){
                if(dependency.right.size() > 1) {
                    for(Integer rightValue : dependency.getRight()) {
                        FunctionalDependency dep = new FunctionalDependency();
                        dep.left = dependency.left;
                        Set<Integer> rightval = new HashSet<Integer>();
                        rightval.add(rightValue);
                        dep.right = rightval;
                        splitFds.add(dep);
                    }
                }else {
                    splitFds.add(dependency);
                }
            }
            return splitFds;
        }

        /**
         * Prints the attributes of a relation
         * @return String with the attributes separated by a coma
         */
        @Override
        public String toString(){
            StringBuilder result = new StringBuilder();

            Iterator<Integer> it = attributes.iterator();
            while(it.hasNext()){
                result.append(it.next());
                result.append(",");
            }
            result.deleteCharAt(result.length()-1);
            return result.toString();
        }

        // For simplicity, equals only looks at the attributes.
        // used for testing only

        /**
         * check if two relations have the same attributes
         * @param other an other relation
         * @return true is the two relations have the same attributes
         */
        @Override
        public boolean equals(Object other){
            if(other == null){
                return false;
            }
            if(! (other instanceof Relation)){
                return false;
            }
            Relation o = (Relation) other;
            return this.attributes.equals(o.attributes);
        }

    }

    static public class FunctionalDependency{
        private Set<Integer> left = new HashSet<>();
        private Set<Integer> right = new HashSet<>();

        /**
         * Default Constructor
         */
        public FunctionalDependency() {

        }

        /**
         * Constructor taking the LHS and the RHS of a functional Dependency
         * @param left LHS
         * @param right RHS
         */
        public FunctionalDependency(Set<Integer> left, Set<Integer> right){
            this.left = left;
            this.right = right;
        }

        /**
         * Constructor taking the LHS and a RHS with only one attribute
         * @param left
         * @param right
         */
        public FunctionalDependency(Set<Integer> left, Integer right){
            this.left = left;
            this.right = new HashSet<>();
            this.right.add(right);
        }


        /**
         * Constructor taking a string representation of the functional dependency
         * @param dp string of the form 1,3;5,6 where LFS = [1,3] and RHS = [5,6]
         */
        public FunctionalDependency(String dp){
            String[] elements = dp.split(";");
            if(elements.length != 2){
                throw new IllegalArgumentException("dependencies need to be of the form AG->CD.\n");
            }
            elements[0] = elements[0].trim().replace(",", "");
            elements[1] = elements[1].trim().replace(",", "");
            for(int i = 0; i < elements[0].length(); i++){
                try{
                    left.add(Integer.parseInt(elements[0].charAt(i)+""));
                } catch (NumberFormatException e) {
                }
            }
            for(int i = 0; i < elements[1].length(); i++){
                try{
                    right.add(Integer.parseInt(elements[1].charAt(i) + ""));
                } catch (NumberFormatException e) {
                    // ignore if not a number, probably a coma
                }
            }
        }

        /**
         * Return the LHS
         * @return LHS
         */
        public Set<Integer> getLeft(){
            return left;
        }

        /**
         * Return the RHS
         * @return RHS
         */
        public Set<Integer> getRight(){
            return right;
        }
    }

}
